
// reference method signatures

vec3        exposure_ref( in vec3 color );
vec3        expo_bright_ref( in vec3 color );
vec3        contrast_ref(   in vec3 color );
vec3        brightness_ref(   in vec3 color );
vec3        shadows_ref(   in vec3 color );
vec3        highlights_ref( in vec3 color );

vec2        blackbody_chromaticity_ref( in float K );
vec3        white_balance_ref(  in vec3 color );
vec3        white_balance_lin( in vec3 color );

// ---------------------------------------------------------------------

// external method signatures

vec3    srgb_to_linear_ref( in vec3 srgb );
vec3    linear_to_srgb_ref( in vec3 linear );

vec3    srgb_to_linear_opt( in vec3 srgb );
vec3    linear_to_srgb_opt( in vec3 linear );

vec3    rgb_to_xyz( in vec3 rgb );
vec3    xyz_to_rgb( in vec3 xyz );

vec3    rgb_to_yiq( in vec3 rgb );
vec3    yiq_to_rgb( in vec3 yiq );

vec3    rgb_to_hsv( in vec3 rgb );
vec3    hsv_to_rgb( in vec3 hsv );

float   min_channel( in vec2 v );
float   min_channel( in vec3 v );
float   min_channel( in vec4 v );
float   max_channel( in vec2 v );
float   max_channel( in vec3 v );
float   max_channel( in vec4 v );

float   sum( in vec2 v );
float   sum( in vec3 v );
float   sum( in vec4 v );

float   mean( in vec2 v );
float   mean( in vec3 v );
float   mean( in vec4 v );

vec2    pow( in vec2 v, in float e );
vec3    pow( in vec3 v, in float e );
vec4    pow( in vec4 v, in float e );

// ---------------------------------------------------------------------

uniform float      u_exposure;
uniform float      u_brightness;
uniform float      u_expoBright;
uniform float      u_shadows;
uniform float      u_highlights;
uniform float      u_contrastGain;
uniform float      u_contrastOffset;

vec3 
exposure_ref( in vec3 color )
{
    return color * vec3( pow( 2., u_exposure ) );
}

vec3
expo_bright_ref( in vec3 color )
{
    float expbri_lin = pow( 2., u_expoBright );

    // behave like normal exposure slider when darkening
    if( u_expoBright <= 0. )
    {
        float brightabs = abs( u_expoBright ) / 2.;
        float brightalpha = .5 * clamp( brightabs, 0., 1. );
        color      = ( 1. - brightalpha ) * color;
        color      = pow( color, .5 * brightabs + 1.);        
    }
    else
    {
        expbri_lin = mix( expbri_lin, 
                          1.5 * expbri_lin, 
                          clamp( 3. * ( expbri_lin - 1. ), 0., 1. ) );
        color =  vec3( 1. ) - pow( ( vec3( 1. ) - color ), expbri_lin );
    }
    
    return color;
}

vec3
sigmoid( in vec3 color, in float gain, in float offset )
{
	return 1. / ( 1. + exp( gain * ( offset - color ) ) );
}

vec3		
sigmoidinv( in vec3 color, in float gain, in float offset, 
            in vec3 nmin, in vec3 nmax )
{
	return offset - log( 1. / ( color * ( nmax - nmin ) + nmin ) - 1. ) / gain;
}

vec3
contrast_ref( in vec3 color )
{
	vec3 nmin = sigmoid( vec3( 0. ), u_contrastGain, u_contrastOffset );
	vec3 nmax = sigmoid( vec3( 1. ), u_contrastGain, u_contrastOffset );

    vec3 nmin2 = sigmoid( vec3( 0. ), u_contrastGain / 2., u_contrastOffset );
	vec3 nmax2 = sigmoid( vec3( 1. ), u_contrastGain / 2., u_contrastOffset );
    
	float gainabs   = abs( u_contrastGain ) + 1.;
    
	vec3 ncolor;
    
	if( u_contrastGain == 0. )
	{
		ncolor = color;
	}
	else
	{
		if( u_contrastGain >= 0. )
		{
			ncolor = sigmoid( color, u_contrastGain, u_contrastOffset );
			ncolor = ( ncolor - nmin ) / ( nmax - nmin );
		}
		else
		{
            ncolor = mix( color, vec3( u_contrastOffset - .5 + .1864 ), -u_contrastGain / 20. );
			ncolor = sigmoidinv( ncolor, u_contrastGain / 2., u_contrastOffset, nmin2, nmax2 );
		}
        
		if( gainabs < 2. )
		{
            float alpha = gainabs - 1.;
            ncolor = mix( color, ncolor, alpha );
		}
	}
    
    return ncolor;
}

vec3
brightness_ref( in vec3 color )
{
    vec3 ncolor = color;

    float brightabs = abs( u_brightness );
    
    if( brightabs > 0. )
    {
        float brightalpha = .01 * clamp( brightabs, 0., 1. );
        if( u_brightness > 0. )
        {
            ncolor      = ( 1. - brightalpha ) * ncolor + brightalpha;
            ncolor      = pow( ncolor, 1. / ( brightabs + 1. ) );
        }
        else
        {
            ncolor      = ( 1. - brightalpha ) * ncolor;
            ncolor      = pow( ncolor, brightabs + 1. );
        }
    }
    
    return ncolor;
}

vec3
shadows_ref( in vec3 color )
{
    vec3 ncolor = color;

    float shadowabs = abs( u_shadows );
    
    if( shadowabs > 0. )
    {
        vec3 shadowalpha = clamp( 4.8 * ncolor - .6, 0., 1. );
        vec3 scolor;
        if( u_shadows > 0. ) {
            scolor = pow( ncolor, 1. / ( shadowabs + 1. ) );
        }
        else {
            scolor = pow( ncolor, shadowabs + 1. );
        }
        ncolor = mix( scolor, color, shadowalpha );
    }
    
	return ncolor;
}

vec3
highlights_ref( in vec3 color )
{
    vec3 ncolor = color;
    
    float highlightabs = abs( u_highlights );
    
    if( highlightabs > 0. )
    {
        ncolor = vec3( 1. ) - ncolor;
        
        vec3 highlightalpha = clamp( 5. * color - 4., 0., 1. );
        vec3 scolor;
        if( u_highlights > 0. ) {
            scolor = pow( ncolor, 1. / ( 2. * highlightabs + 1. ) );
        }
        else {
            scolor = pow( ncolor, 2. * highlightabs + 1. );
        }
        ncolor = mix( color, 1. - scolor, highlightalpha );
    }
    
	return ncolor;
}

// ---------------------------------------------------------------------

uniform vec3    u_whiteBalanceColor;
uniform float   u_colorTemperature;

vec2
blackbody_chromaticity( in float K )
{
	if( ( K < 1667. ) || ( K > 25000. ) )
    return vec2( 0. );
    
	vec2 xy;
    
    if( K < 4000. )
        xy.x = -0.2661239e9 / pow( K, 3. ) - 0.2343580e6 / pow( K, 2. ) + 0.8776956e3 / K + 0.179910;
    else
        xy.x = -3.0258469e9 / pow( K, 3. ) + 2.1070379e6 / pow( K, 2. ) + 0.2226347e3 / K + 0.240390;
    
	if( K < 2222. )
        xy.y = -1.1063814  * pow( xy.x, 3. ) - 1.34811020 * pow( xy.x, 2. ) + 2.18555832 * xy.x - 0.20219683;
	else {
        if( K < 4000. ) {
            xy.y = -0.9549476  * pow( xy.x, 3. ) - 1.37418593 * pow( xy.x, 2. ) + 2.09137015 * xy.x - 0.16748867;
        } else {
            xy.y = +3.0817580  * pow( xy.x, 3. ) - 5.87338670 * pow( xy.x, 2. ) + 3.75112997 * xy.x - 0.37001483;
        }
    }
    
	return xy;
}

vec3
white_balance_ref( in vec3 color )
{
    mat3 XYZ_Minv 	= mat3( 	 3.2404542, -1.5371385, -0.4985314,
                                -0.9692660,  1.8760108,  0.0415560,
                                 0.0556434, -0.2040259,  1.0572252 );
    
	vec2 dest_xy	= blackbody_chromaticity( u_colorTemperature );
    
	vec3 dest_xyz;
	dest_xyz.x = dest_xy.x / dest_xy.y;
	dest_xyz.y = 1.;
	dest_xyz.z = ( 1. - dest_xy.x - dest_xy.y ) / dest_xy.y;
    
	vec3 dest_rgb	= dest_xyz * XYZ_Minv;
    
	return color * dest_rgb;
}

vec3
white_balance_lin( in vec3 color )
{
    color = color * u_whiteBalanceColor;
    return color;
}
