
precision mediump float;

// ---------------------------------------------------------------------

varying mediump vec2    v_iTexCoord;
varying mediump vec2    v_aTexCoord;

uniform lowp sampler2D  u_image;
uniform lowp sampler2D  u_alphaMask;

// ---------------------------------------------------------------------

// optimized method signatures

vec3    luminance_lut( in vec3 color );

vec3    vibrance( in vec3 color );
vec3    white_balance_srgb( in vec3 color );
vec3    color_adjust( in vec3 color );

float   greyscale_blend( in vec3 color );
vec3    greyscale_tint( in float grey );
vec3    greyscale_adjust( in vec3 color );
float   greyscale_lum( in vec3 color );

float   alpha_adjust( in float alpha );

vec3    final_mix( in vec3 color, in vec3 grey, in float alpha );
vec3    ruby_mix(  in vec3 grey, in float alpha );

// reference method signatures

vec3    exposure_ref( in vec3 color );
vec3    expo_bright_ref( in vec3 color );
vec3    contrast_ref(   in vec3 color );
vec3    brightness_ref(   in vec3 color );
vec3    shadows_ref(   in vec3 color );
vec3    highlights_ref( in vec3 color );

vec2    blackbody_chromaticity_ref( in float K );
vec3    white_balance_ref(  in vec3 color );
vec3    white_balance_lin( in vec3 color );

// other method signatures

vec3    srgb_to_linear_ref( in vec3 srgb );
vec3    linear_to_srgb_ref( in vec3 linear );

float   mean( in vec2 v );
float   mean( in vec3 v );
float   mean( in vec4 v );

// ---------------------------------------------------------------------

vec3        process_ref( in vec3 color, in float alpha );
lowp vec3   process_opt( in lowp vec3 color, in lowp float alpha );
vec3        process_diff( in vec3 color, in float alpha );

// ---------------------------------------------------------------------

void main()
{    
    vec3 color          = texture2D( u_image, v_iTexCoord ).rgb;
    float alpha         = texture2D( u_alphaMask,  v_aTexCoord ).a; 
        
#ifndef PP_IMAGE_COMPARE
    
    color = process_ref( color, alpha );
    
#else

    color = process_diff( color, alpha );

#endif
    
    gl_FragColor    = vec4( color, 1. );
}

// ---------------------------------------------------------------------

vec3
process_ref( in vec3 color, in float alpha )
{
    // luminance adjustment
    color       = srgb_to_linear_ref( color );
    color       = expo_bright_ref( color );
    color       = contrast_ref( color );
    color       = linear_to_srgb_ref( color );
    color       = shadows_ref( color );

    // blend and tint grey
    vec3 grey   = greyscale_adjust( color );
    
    // adjust saturation and white balance
    color       = vibrance( color );
    color       = srgb_to_linear_ref( color );
    color       = white_balance_lin( color );
//    color       = white_balance_ref( color );
    color       = linear_to_srgb_ref( color );
//    color       = white_balance_srgb( color );
    
    // mix color and grey
    color       = final_mix( color, grey, alpha );
    
    return color;
}

lowp vec3
process_opt( in lowp vec3 color, in lowp float alpha )
{
    // luminance lut
    color       = luminance_lut( color );
    
    // blend and tint grey
    lowp vec3 grey   = greyscale_adjust( color );
    
    // adjust saturation and white balance
    color       = color_adjust( color );
    
    // mix color and grey
    color       = final_mix( color, grey, alpha );
    
    return color;
}

vec3
process_diff( in vec3 color, in float alpha )
{
    // compute reference color
    vec3 ref_color  = process_ref( color, alpha );

    // compute optimized color
    vec3 opt_color  = process_opt( color, alpha );
    
    // compute difference
    color           = vec3( 10. * distance( ref_color, opt_color ) );
    
    return color;
}
