precision mediump float;

// ---------------------------------------------------------------------

varying highp vec2      v_tTexCoord;

uniform highp float       u_pxStart;

uniform highp vec2   u_dimsTile;
uniform highp vec2   u_dimsImage;
uniform highp vec2    u_maxAlphaCoord;

uniform lowp int        u_orientFlag;

uniform lowp sampler2D  u_image;
uniform lowp sampler2D  u_alphaMask;

// ---------------------------------------------------------------------

vec3    luminance_lut( in vec3 color );

vec3    vibrance( in vec3 color );
vec3    white_balance_srgb( in vec3 color );
vec3    color_adjust( in vec3 color );

float   greyscale_blend( in vec3 color );
vec3    greyscale_tint( in float grey );
vec3    greyscale_adjust( in vec3 color );
float   greyscale_lum( in vec3 color );

float   alpha_adjust( in float alpha );

vec3    final_mix( in vec3 color, in vec3 grey, in float alpha );
vec3    ruby_mix(  in vec3 grey, in float alpha );

// ---------------------------------------------------------------------

mediump vec2 orient_alpha( in mediump vec2 aTexCoord, in lowp int orientFlag );

lowp vec3 process_norm( in lowp vec3 color, in lowp float alpha );

// ---------------------------------------------------------------------

void main()
{
    // compute the index of the pixel if the image was a 1-D array
    highp float pxOffset  = u_pxStart
                        + floor( v_tTexCoord.y * u_dimsTile.y ) * u_dimsTile.x
                        + floor( v_tTexCoord.x * u_dimsTile.x );
                            
    // re-map that index into the 2D array of the original image
    mediump vec2 iTexCoordU   = vec2(   mod(   pxOffset,  u_dimsImage.x ),
                                        floor( pxOffset / u_dimsImage.x ) );

    // convert to the coordinate system of the alpha texture
    // we add (0.5, 0.5) because textures are sampled in the pixel/texel center
    mediump vec2 aTexCoord    = (iTexCoordU + vec2(0.5, 0.5))  * (u_maxAlphaCoord / u_dimsImage);
    aTexCoord                 = orient_alpha( aTexCoord, u_orientFlag );

    vec3 color          = texture2D( u_image,       v_tTexCoord ).rgb;
    lowp float alpha    = texture2D( u_alphaMask,   aTexCoord ).a;
        
    color               = process_norm( color, alpha );
    
    gl_FragColor        = vec4( color, 1. );
}

// ---------------------------------------------------------------------

mediump vec2 orient_alpha( in mediump vec2 aTexCoord, in lowp int orientFlag )
{
    //    // UIImageOrientationUp
    //    if( u_orientFlag == 0 )
    //    { }

    // UIImageOrientationDown
    if( u_orientFlag == 1 )
        aTexCoord   = u_maxAlphaCoord - aTexCoord;

    // UIImageOrientationLeft
    if( u_orientFlag == 2 )
        aTexCoord   = vec2( aTexCoord.y, u_maxAlphaCoord.x - aTexCoord.x );

    // UIImageOrientationRight
    if( u_orientFlag == 3 )
        aTexCoord   = vec2( u_maxAlphaCoord.y - aTexCoord.y, aTexCoord.x );

    // UIImageOrientationUpMirrored
    if( u_orientFlag == 4 )
        aTexCoord.x = u_maxAlphaCoord.x - aTexCoord.x;

    // UIImageOrientationDownMirrored
    if( u_orientFlag == 5 )
        aTexCoord.y = u_maxAlphaCoord.y - aTexCoord.y;

    // UIImageOrientationLeftMirrored
    if( u_orientFlag == 6 )
        aTexCoord   = aTexCoord.yx;

    // UIImageOrientationRightMirrored
    if( u_orientFlag == 7 )
        aTexCoord   = ( u_maxAlphaCoord - aTexCoord ).yx;

    return aTexCoord;
}

lowp vec3
process_norm( in lowp vec3 color, in lowp float alpha )
{
    // luminance lut
    mediump vec3 adjusted_color = luminance_lut( color );
    
    // blend and tint grey
    lowp vec3 grey  = greyscale_adjust( adjusted_color );
    
    // adjust saturation and white balance
    adjusted_color = color_adjust( adjusted_color );
    
    // mix color and grey
    lowp vec3 final_color = final_mix( adjusted_color, grey, alpha );
    
    return final_color;
}
